import hashlib

class ResponseError(Exception):
    pass

class Response:
    status: int = 10
    message: str = ""
    content_type: str = ""
    len: int
    hash: str

    keys: dict[str, str]
    content: str

    def __init__(self, status: int=0, message: str="", content_type: str="text/molerat", keys: dict[str, str]={}, content: str=""):
        if status: self.status = status
        if message: self.message = message
        if content_type: self.content_type = content_type
        
        if keys and content:
            raise ResponseError("Keys and data cannot coexist in the same response")

        self.keys = keys
        self.content = content
    
    def stringify(self) -> str:
        string = ""

        string += f"{self.status}\r\n"

        content = ""
        if self.message:
            string += f"message:{self.message}\t\r\n"

        if self.keys:
            for key in self.keys:
                content += f"{key}:{self.keys[key]}\t\r\n"

        if self.content:
            content = self.content
            

        if content:
            if self.content_type:
                string += f"type:{self.content_type}\t\r\n"
            else:
                raise ResponseError("Content defined without content_type")

            string += f"length:{len(content)}\t\r\n"
            string += f"hash:{hashlib.sha256(content.encode()).hexdigest()}\t\r\n"
            string += "\r\n\r\n"
            string += content
        else:
            string += "\r\n\r\n"

        return string

    def __str__(self):
        return self.stringify()



import os
import socket

from OpenSSL import SSL
from typing import Callable

from moleserv.response import Response
from moleserv.request import Request, RequestMethod, RequestParseError

class PathNotFoundError(Exception):
    pass

class Server:
    keyfile: str
    certfile: str

    address: str
    port: int

    _get_paths: dict[str, Callable[[Request], Response]] = {}
    _put_paths: dict[str, Callable[[Request], Response]] = {}
    _del_paths: dict[str, Callable[[Request], Response]] = {}

    def __init__(self, address: str, port: int=2693):
        self.address = address
        self.port = port
    
    def get(self, path: str, callback: Callable[[Request], Response]):
        self._get_paths[path] = callback
        return self

    def put(self, path: str, callback: Callable[[Request], Response]):
        self._put_paths[path] = callback
        return self

    def delete(self, path: str, callback: Callable[[Request], Response]):
        self._del_paths[path] = callback
        return self

    def listen(self, keyfile: str, certfile: str):
        self.keyfile = keyfile
        self.certfile = certfile

        ctx = SSL.Context(SSL.SSLv23_METHOD)
        ctx.use_privatekey_file(self.keyfile)
        ctx.use_certificate_file(self.certfile)

        server = SSL.Connection(ctx, socket.socket(socket.AF_INET, socket.SOCK_STREAM))
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind(("127.0.0.1", 2693))
        server.listen(15)

        while True:
            conn, addr = server.accept()
            print(f"New connection from '{addr}'")

            try:
                self._handle_connection(conn)
                print(f"Closing connection from '{addr}'")
            except RequestParseError as err:
                res = Response(err.code, "Malformed request")
                conn.send(res.stringify().encode())
                print(f"Closing connection from '{addr}' with error '{err}' and code '{err.code}'")
            except PathNotFoundError as err:
                res = Response(32, "Not found")
                conn.send(res.stringify().encode())
                print(f"Closing connection from '{addr}' with error '{err}' and code '32'")
            except Exception as err:
                res = Response(40, "Internal error")
                conn.send(res.stringify().encode())
                print(f"Closing connection from '{addr}' with error '{err}' and code '40'")

            conn.shutdown()
            conn.sock_shutdown(socket.SHUT_RDWR)
            conn.close()

    def _handle_connection(self, conn):
        total_data = ""
        total_bytes_read = 0

        data = conn.recv(2048).decode()
        total_data += data
        total_bytes_read += len(data)

        sections = data.split("\r\n\r\n")
        if len(sections) < 2:
            raise Exception("No headers, or headers too long")

        req = Request(data)


        while total_bytes_read < req.length:
            data = conn.recv(total_bytes_read).decode()
            total_data += data
            total_bytes_read += len(data)

            req = Request(total_data)


        res = Response()
        match req.kind:
            case RequestMethod.GET:
                if req.url.path in self._get_paths:
                    res = self._get_paths[req.url.path](req)
                else:
                    raise PathNotFoundError(f"Path '{req.url.path}' not found")
            case RequestMethod.PUT:
                if req.url.path in self._put_paths:
                    res = self._put_paths[req.url.path](req)
                else:
                    raise PathNotFoundError(f"Path '{req.url.path}' not found")
            case RequestMethod.DEL:
                if req.url.path in self._del_paths:
                    res = self._del_paths[req.url.path](req)
                else:
                    raise PathNotFoundError(f"Path '{req.url.path}' not found")

        conn.send(res.stringify().encode())



from enum import Enum
from urllib.parse import urlparse, ParseResult

class RequestParseError(Exception):
    def __init__(self, message, code: int = 40):            
        super().__init__(message)
        self.code = code

class RequestMethod(Enum):
    GET=0
    PUT=1
    DEL=2

class Request:
    kind: RequestMethod
    url: ParseResult

    length: int = 0
    hash: str = ""

    keys: dict[str, str] = {}

    def __init__(self, data):
        self.parse(data)
    
    def parse(self, data):
        self._parse_head(data)
        match self.kind:
            case RequestMethod.GET:
                pass # get is very simple
            case RequestMethod.PUT:
                self._parse_put(data)
            case RequestMethod.DEL:
                self._parse_del(data)

    def _parse_head(self, data):
        components = data.split("\r\n")[0].split(" ")
        if len(components) != 2:
            raise RequestParseError("Head line has more than one part", 30)

        match components[0]:
            case "get":
                self.kind = RequestMethod.GET
            case "put":
                self.kind = RequestMethod.PUT
            case "del":
                self.kind = RequestMethod.DEL

            case _:
                raise RequestParseError("No request headers", 30)

        self.url = urlparse(f"molerat://{components[1].strip()}")

    def _parse_put(self, data):
        fields = data.split("\r\n\r\n")[0].split("\r\n")[1:]
        fields = "\r\n".join(fields)

        kv = self.parse_kv(fields)

        for key, val in kv.items():
            match key:
                case "length":
                    self.length = int(val)
                case "hash":
                    self.hash = val
                case _:
                    raise RequestParseError("Unrecognized header key", 30)

        if (self.hash and not self.length) or (self.length and not self.hash):
            raise RequestParseError("Hash and length must both be specified", 30)

        fields = data.split("\r\n\r\n")[1].split("\r\n")
        fields = "\r\n".join(fields)

        self.keys = self.parse_kv(fields)


    def _parse_del(self, data):
        self._parse_put(data) # dels are the same as puts

    def parse_kv(self, data: str) -> dict[str, str]:
        fields = data.split("\t\r\n")

        kv = {}
        for field in fields:
            key_val = field.split(":")
            key = key_val[0]
            val = ":".join(key_val[1:])
            kv[key] = val

        return kv
        

    def __str__(self):
        string = "Request(\r\n\t"
        string += f"{self.kind}\r\n\t"
        string += f"{self.url}\r\n\t"

        string += f"{self.length}\r\n\t"
        string += f"{self.hash}\r\n\t"

        string += f"{self.keys}\r\n\t"
        string += "\r)"

        return string




from setuptools import find_packages, setup

setup(
        name="moleserv",
        version="0.1.2",
        description="Molerat protocol server library.",
        url="https://git.trinket.icu/moleserv.git",

        long_description=open("README.md", 'r').read(),
        long_description_content_type='text/markdown',

        author="Shav Kinderlehrer",
        author_email="molerat@git.trinket.icu",

        license="0BSD",

        packages=find_packages(include=["moleserv"]),

        install_requires=["pyOpenSSL"],
        python_requires=">=3.11",

        project_urls={
            "Source": "https://git.trinket.icu/moleserv.git",
            "Github": "https://github.com/secondary-smiles/moleserv",
        },

        keywords="network server library molerat protocol",
        classifiers=[
            "License :: OSI Approved :: BSD License",

            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.11",

            "Topic :: Communications",
            "Topic :: Communications :: File Sharing",
            "Topic :: Internet",
            "Topic :: Software Development :: Libraries",
        ],
)
